# coding:utf-8

'''
Created on 27.03.2010

    @license: GNU GPL v3 or above, see LICENSE for more details.
    @copyleft: 2010 by the django-weave team, see AUTHORS for more details.
'''
# Due to Mozilla Weave supporting Recaptcha solely, we have to stick with it until
# they decide to change the interface to pluggable captchas.
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.contrib.csrf.middleware import csrf_exempt
from django.http import HttpResponse

# django-weave own stuff
from weave import Logging
from weave.utils import assert_weave_version

logger = Logging.get_logger()

@csrf_exempt
def captcha(request, version):
    assert_weave_version(version)
    # Check for aviability of recaptcha 
    # (can be found at: http://pypi.python.org/pypi/recaptcha-client)
    try:
        from recaptcha.client.captcha import displayhtml
    except ImportError:
        logger.error("Captcha requested but unable to import the 'recaptcha' package!")
        return HttpResponse("Captcha support disabled due to missing Python package 'recaptcha'.")
    if not getattr(settings, "RECAPTCHA_PUBLIC_KEY"):
        logger.error("Trying to create user but settings.RECAPTCHA_PUBLIC_KEY not set")
        raise ImproperlyConfigured
    # Send a simple HTML to the client. It get's rendered inside the Weave client.
    return HttpResponse("<html><body>%s</body></html>" % displayhtml(settings.RECAPTCHA_PUBLIC_KEY))



#!/usr/bin/env python
# coding: utf-8

"""
    weave - manage.py
    ~~~~~~~~~~~~~~~~~~~
    
    http://docs.djangoproject.com/en/dev/ref/django-admin/
    
    borrowed from the pinax project.
"""

import os
import sys

if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = "testproject.settings"
    print "DJANGO_SETTINGS_MODULE not set, use: %r" % os.environ['DJANGO_SETTINGS_MODULE']

def _error(msg):
    print "Import Error:", msg
    print "-" * 79
    import traceback
    traceback.print_exc()
    print "-" * 79
    import sys
    for p in sys.path:
        print p
    print "-" * 79
    print "Did you activate the virtualenv?"
    sys.exit(1)

try:
    from django.core.management import setup_environ, execute_from_command_line
except ImportError, msg:
    _error(msg)


try:
    import weave
except ImportError, msg:
    _error(msg)
except:
    import traceback
    traceback.print_exc()


try:
    import settings as settings_mod # Assumed to be in the same directory.
except ImportError:
    import sys
    sys.stderr.write("Error: Can't find the file 'settings.py' in the directory containing %r. It appears you've customized things.\nYou'll have to run django-admin.py, passing it your settings module.\n(If the file settings.py does indeed exist, it's causing an ImportError somehow.)\n" % __file__)
    sys.exit(1)
except:
    import traceback
    traceback.print_exc()

try:
    # setup the environment before we start accessing things in the settings.
    setup_environ(settings_mod)
except:
    import traceback
    traceback.print_exc()

if __name__ == "__main__":
    try:
        execute_from_command_line()
    except Exception:
        import traceback
        traceback.print_exc()




"""
    weave_server.py [options]

    This is a simple reference implementation for a Weave server,
    which can also be used to test the Weave client against.

    This server behaves like a standard Weave server, with the
    following limitations and/or enhancements:

      * Responses to CAPTCHA challenges during new user registration
        are always accepted, with the exception that the magic word
        'bad' always fails (this was added for testing purposes).

      * Validation emails are not sent, though the server tells
        clients that they are.

      * The url at '/state/' can be accessed to retrieve a snapshot of
        the current state of the server.  This data can be saved to a
        file and restored later with the '--state' command-line
        option.  The state is also just a Pythonic representation of
        the server's state, and as such is relatively human-readable
        and can be hand-edited if necessary.

      * Timeout values sent with HTTP LOCK requests are ignored.

      * The server operates over HTTP, not HTTPS.
"""

from wsgiref.simple_server import make_server
from optparse import OptionParser
import httplib
import base64
import logging
import pprint
import cgi

DEFAULT_PORT = 8000
DEFAULT_REALM = "services.mozilla.com"
CAPTCHA_FAILURE_MAGIC_WORD = "bad"

class HttpResponse(object):
    def __init__(self, code, content = "", content_type = "text/plain"):
        self.status = "%s %s" % (code, httplib.responses.get(code, ""))
        self.headers = [("Content-type", content_type)]
        if code == httplib.UNAUTHORIZED:
            self.headers +=  [("WWW-Authenticate",
                               "Basic realm=\"%s\"" % DEFAULT_REALM)]
        if not content:
            content = self.status
        self.content = content

class HttpRequest(object):
    def __init__(self, environ):
        self.environ = environ
        content_length = environ.get("CONTENT_LENGTH")
        if content_length:
            stream = environ["wsgi.input"]
            self.contents = stream.read(int(content_length))
        else:
            self.contents = ""

class Perms(object):
    # Special identifier to indicate 'everyone' instead of a
    # particular user.
    EVERYONE = 0

    def __init__(self, readers=None, writers=None):
        if not readers:
            readers = []
        if not writers:
            writers = []

        self.readers = readers
        self.writers = writers

    def __is_privileged(self, user, access_list):
        return (user in access_list or self.EVERYONE in access_list)

    def can_read(self, user):
        return self.__is_privileged(user, self.readers)

    def can_write(self, user):
        return self.__is_privileged(user, self.writers)

    def __acl_repr(self, acl):
        items = []
        for item in acl:
            if item == self.EVERYONE:
                items.append("Perms.EVERYONE")
            else:
                items.append(repr(item))
        return "[" + ", ".join(items) + "]"

    def __repr__(self):
        return "Perms(readers=%s, writers=%s)" % (
            self.__acl_repr(self.readers),
            self.__acl_repr(self.writers)
            )

def requires_read_access(function):
    function._requires_read_access = True
    return function

def requires_write_access(function):
    function._requires_write_access = True
    return function

class WeaveApp(object):
    """
    WSGI app for the Weave server.
    """

    __CAPTCHA_HTML = '<html><head></head><body><script type=\'text/javascript\'>var RecaptchaOptions = {theme: \'red\', lang: \'en\'};</script><script type="text/javascript" src="http://api.recaptcha.net/challenge?k=6Lc_HwIAAAAAACneEwAadA-wKZCOrjo36TFQv160"></script>\n\n\t<noscript>\n  \t\t<iframe src="http://api.recaptcha.net/noscript?k=6Lc_HwIAAAAAACneEwAadA-wKZCOrjo36TFQv160" height="300" width="500" frameborder="0"></iframe><br/>\n  \t\t<textarea name="recaptcha_challenge_field" rows="3" cols="40"></textarea>\n  \t\t<input type="hidden" name="recaptcha_response_field" value="manual_challenge"/>\n\t</noscript></body></html>'

    def __init__(self, state=None):
        self.contents = {}
        self.dir_perms = {"/" : Perms(readers=[Perms.EVERYONE])}
        self.passwords = {}
        self.email = {}
        self.locks = {}
        self._tokenIds = 0

        if state:
            self.__setstate__(state)

    def add_user(self, username, password, email = None):
        assert username, "Username cannot be empty"
        assert password, "Password cannot be empty"

        home_dir = "/user/%s/" % username
        public_dir = home_dir + "public/"
        self.dir_perms[home_dir] = Perms(readers=[username],
                                         writers=[username])
        self.dir_perms[public_dir] = Perms(readers=[Perms.EVERYONE],
                                           writers=[username])
        self.passwords[username] = password
        if email:
            self.email[email] = username

    def __get_perms_for_path(self, path):
        possible_perms = [dirname for dirname in self.dir_perms
                          if path.startswith(dirname)]
        possible_perms.sort(key = len)
        perms = possible_perms[-1]
        return self.dir_perms[perms]

    def __get_files_in_dir(self, path):
        return [filename for filename in self.contents
                if filename.startswith(path)]

    def __api_share(self, path):
        params = cgi.parse_qs(self.request.contents)
        user = params["uid"][0]
        password = params["password"][0]
        if self.passwords.get(user) != password:
            return HttpResponse(httplib.UNAUTHORIZED)
        else:
            import json
            cmd = json.read(params["cmd"][0])
            dirname = "/user/%s/%s" % (user, cmd["directory"])
            if not dirname.endswith("/"):
                dirname += "/"
            readers = []
            for reader in cmd["share_to_users"]:
                if reader == "all":
                    readers.append(Perms.EVERYONE)
                else:
                    readers.append(reader)
            if user not in readers:
                readers.append(user)
            self.dir_perms[dirname] = Perms(readers = readers,
                                            writers = [user])
            return HttpResponse(httplib.OK, "OK")
    
    # Registration API
    def __api_register_check(self, what, where):
        what = what.strip("/")
        if what.strip() == "":
            return HttpResponse(400,
                                self.ERR_WRONG_HTTP_METHOD)
            
        if what in where:
            return HttpResponse(httplib.OK,
                                self.ERR_UID_OR_EMAIL_IN_USE)
        else:
            return HttpResponse(httplib.OK,
                                self.ERR_UID_OR_EMAIL_AVAILABLE)

    ERR_UID_OR_EMAIL_AVAILABLE = "1"
    ERR_WRONG_HTTP_METHOD = "-1"
    ERR_MISSING_UID = "-2"
    ERR_INVALID_UID = "-3"
    ERR_UID_OR_EMAIL_IN_USE = "0"
    ERR_EMAIL_IN_USE = "-5"
    ERR_MISSING_PASSWORD = "-8"
    ERR_MISSING_RECAPTCHA_CHALLENGE_FIELD = "-6"
    ERR_MISSING_RECAPTCHA_RESPONSE_FIELD = "-7"
    ERR_MISSING_NEW = "-11"
    ERR_INCORRECT_PASSWORD = "-12"
    ERR_ACCOUNT_CREATED_VERIFICATION_SENT = "2"
    ERR_ACCOUNT_CREATED = "3"

    __REQUIRED_CHANGE_PASSWORD_FIELDS = ["uid", "password", "new"]

    __REQUIRED_NEW_ACCOUNT_FIELDS = ["uid",
                                     "password",
                                     "recaptcha_challenge_field",
                                     "recaptcha_response_field"]

    __FIELD_ERRORS = {
        "uid" : ERR_MISSING_UID,
        "password" : ERR_MISSING_PASSWORD,
        "new" : ERR_MISSING_NEW,
        "recaptcha_challenge_field" : ERR_MISSING_RECAPTCHA_CHALLENGE_FIELD,
        "recaptcha_response_field" : ERR_MISSING_RECAPTCHA_RESPONSE_FIELD
        }

    def __get_fields(self, required_fields):
        params = cgi.parse_qs(self.request.contents)
        fields = {}
        for name in params:
            fields[name] = params[name][0]
        for name in required_fields:
            if not fields.get(name):
                return HttpResponse(httplib.BAD_REQUEST,
                                    self.__FIELD_ERRORS[name])
        return fields

    def __api_create_account(self, path):
        fields = self.__get_fields(self.__REQUIRED_NEW_ACCOUNT_FIELDS)
        if isinstance(fields, HttpResponse):
            return fields
        if fields["uid"] in self.passwords:
            return HttpResponse(httplib.BAD_REQUEST,
                                self.ERR_UID_OR_EMAIL_IN_USE)
        if fields["recaptcha_response_field"] == CAPTCHA_FAILURE_MAGIC_WORD:
            return HttpResponse(httplib.EXPECTATION_FAILED)
        if fields.get("mail"):
            if self.email.get(fields["mail"]):
                return HttpResponse(httplib.BAD_REQUEST,
                                    self.ERR_EMAIL_IN_USE)
            # TODO: We're not actually sending an email...
            body_code = self.ERR_ACCOUNT_CREATED_VERIFICATION_SENT
        else:
            body_code = self.ERR_ACCOUNT_CREATED

        self.add_user(fields["uid"], fields["password"],
                      fields.get("mail"))
        return HttpResponse(httplib.CREATED, body_code)

    def __api_change_password(self, path):
        fields = self.__get_fields(self.__REQUIRED_CHANGE_PASSWORD_FIELDS)
        if isinstance(fields, HttpResponse):
            return fields
        if not self.passwords.get(fields["uid"]):
            return HttpResponse(httplib.BAD_REQUEST,
                                self.ERR_INVALID_UID)
        if self.passwords[fields["uid"]] != fields["password"]:
            return HttpResponse(httplib.BAD_REQUEST,
                                self.ERR_INCORRECT_PASSWORD)
        self.passwords[fields["uid"]] = fields["new"]
        return HttpResponse(httplib.OK)

    # HTTP method handlers

    @requires_write_access
    def _handle_LOCK(self, path):
        if path in self.locks:
            return HttpResponse(httplib.LOCKED)
        token = "opaquelocktoken:%d" % self._tokenIds
        self._tokenIds += 1
        self.locks[path] = token
        response = """<?xml version="1.0" encoding="utf-8"?>
                   <D:prop xmlns:D="DAV:">
                     <D:lockdiscovery>
                       <D:activelock>
                         <D:locktoken>
                           <D:href>%s</D:href>
                         </D:locktoken>
                       </D:activelock>
                     </D:lockdiscovery>
                   </D:prop>""" % token
        return HttpResponse(httplib.OK, response, content_type="text/xml")

    @requires_write_access
    def _handle_UNLOCK(self, path):
        token = self.request.environ["HTTP_LOCK_TOKEN"]
        if path not in self.locks:
            return HttpResponse(httplib.BAD_REQUEST)
        if token == "<%s>" % self.locks[path]:
            del self.locks[path]
            return HttpResponse(httplib.NO_CONTENT)
        return HttpResponse(httplib.BAD_REQUEST)

    @requires_write_access
    def _handle_MKCOL(self, path):
        return HttpResponse(httplib.OK)

    @requires_write_access
    def _handle_PUT(self, path):
        self.contents[path] = self.request.contents
        return HttpResponse(httplib.OK)

    def _handle_POST(self, path):
        if path == "/api/share/":
            return self.__api_share(path)
        elif path == "/api/register/new/":
            return self.__api_create_account(path)
        elif path == "/api/register/chpwd/":
            return self.__api_change_password(path)
        else:
            return HttpResponse(httplib.NOT_FOUND)

    @requires_write_access
    def _handle_PROPFIND(self, path):
        response = """<?xml version="1.0" encoding="utf-8"?>
                   <D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">"""

        path_template = """<D:response>
                           <D:href>%(href)s</D:href>
                           <D:propstat>
                           <D:prop>
                           %(props)s
                           </D:prop>
                           <D:status>HTTP/1.1 200 OK</D:status>
                           </D:propstat>
                           </D:response>"""

        if path in self.locks:
            props = "<D:locktoken><D:href>%s</D:href></D:locktoken>" % (
                self.locks[path]
                )
        else:
            props = ""

        response += path_template % {"href": path,
                                     "props": props}

        if path.endswith("/"):
            for filename in self.__get_files_in_dir(path):
                response += path_template % {"href" : filename,
                                             "props" : ""}

        response += """</D:multistatus>"""
        return HttpResponse(httplib.MULTI_STATUS, response,
                            content_type="text/xml")

    @requires_write_access
    def _handle_DELETE(self, path):
        response = HttpResponse(httplib.OK)
        if path.endswith("/"):
            # Delete a directory.
            for filename in self.__get_files_in_dir(path):
                del self.contents[filename]
        else:
            # Delete a file.
            if path not in self.contents:
                response = HttpResponse(httplib.NOT_FOUND)
            else:
                del self.contents[path]
        return response

    @requires_read_access
    def _handle_GET(self, path):
        if path in self.contents:
            return HttpResponse(httplib.OK, self.contents[path])
        elif path == "/state/":
            state_str = pprint.pformat(self.__getstate__())
            return HttpResponse(httplib.OK, state_str)
        elif path == "/api/register/new/":
            return HttpResponse(httplib.OK, self.__CAPTCHA_HTML,
                                content_type = "text/html")
        elif path.startswith("/api/register/check/"):
            return self.__api_register_check(path[20:], self.passwords)
        elif path.startswith("/api/register/chkmail/"):
            return self.__api_register_check(path[22:], self.email)
        elif path.endswith("/"):
            return self.__show_index(path)
        else:
            return HttpResponse(httplib.NOT_FOUND)

    def __getstate__(self):
        state = {}
        state.update(self.__dict__)
        del state["request"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __show_index(self, path):
        output = []
        for filename in self.__get_files_in_dir(path):
            output.append("<p><a href=\"%s\">%s</a></p>" % (filename,
                                                            filename))
        if output:
            output = "".join(output)
        else:
            output = ("<p>There are no files under the "
                      "directory <tt>%s</tt>.</p>" % (path))
        return HttpResponse(httplib.OK, output, content_type="text/html")

    def __process_handler(self, handler):
        response = None
        auth = self.request.environ.get("HTTP_AUTHORIZATION")
        if auth:
            user, password = base64.b64decode(auth.split()[1]).split(":")
            if self.passwords.get(user) != password:
                response = HttpResponse(httplib.UNAUTHORIZED)
        else:
            user = Perms.EVERYONE

        if response is None:
            path = self.request.environ["PATH_INFO"]
            perms = self.__get_perms_for_path(path)
            checks = []
            if hasattr(handler, "_requires_read_access"):
                checks.append(perms.can_read)
            if hasattr(handler, "_requires_write_access"):
                checks.append(perms.can_write)
            for check in checks:
                if not check(user):
                    response = HttpResponse(httplib.UNAUTHORIZED)

        if response is None:
            response = handler(path)

        return response

    def __call__(self, environ, start_response):
        """
        Main WSGI application method.
        """

        self.request = HttpRequest(environ)
        method = "_handle_%s" % environ["REQUEST_METHOD"]

        # See if we have a method called 'handle_<method>', where
        # <method> is the name of the HTTP method to call.  If we do,
        # then call it.
        if hasattr(self, method):
            handler = getattr(self, method)
            response = self.__process_handler(handler)
        else:
            response = HttpResponse(
                httplib.METHOD_NOT_ALLOWED,
                "Method %s is not yet implemented." % method
                )

        start_response(response.status, response.headers)
        return [response.content]

if __name__ == "__main__":
    usage = __import__("__main__").__doc__
    parser = OptionParser(usage = usage)
    parser.add_option("-s", "--state", dest="state_filename",
                      help="retrieve server state from filename")
    options, args = parser.parse_args()

    print "Weave Development Server"
    print
    print "Run this script with '-h' for usage information."

    logging.basicConfig(level=logging.DEBUG)

    if options.state_filename:
        filename = options.state_filename
        logging.info("Setting initial state from '%s'." % filename)
        data = open(filename, "r").read()
        state = eval(data)
        app = WeaveApp(state)
    else:
        app = WeaveApp()

    logging.info("Serving on port %d." % DEFAULT_PORT)
    httpd = make_server('', DEFAULT_PORT, app)
    httpd.serve_forever()



'''
Models.
FIXME: I dropped site framework integration for the sake 
       of simpler debugging. Add it back once API is stable 
       and working. 

Created on 15.03.2010

@license: GNU GPL v3 or above, see LICENSE for more details.
@copyright: 2010 see AUTHORS for more details.
@author: Jens Diemer
@author: FladischerMichael
'''

from django.conf import settings
from django.db import models
from django.core.exceptions import ValidationError
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.contrib.sites.managers import CurrentSiteManager

from weave.utils import weave_timestamp


class BaseModel(models.Model):
    modified = models.DateTimeField(auto_now=True, help_text="Time of the last change.")

    class Meta:
        abstract = True


class CollectionManager(CurrentSiteManager):
    def create_or_update(self, user, col_name, timestamp, since=None):
        collection, created = super(CollectionManager, self).get_or_create(
            user=user, name=col_name,
        )

        # See if we have a constraint on the last modified date  
        if since is not None:
            if since < collection.modified:
                raise ValidationError

        collection.modified = timestamp
        collection.save()
        return collection, created


class Collection(BaseModel):
    """
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/Setup
    
    inherited from BaseModel:
        modified -> datetime of the last change
    """
    user = models.ForeignKey(User)
    name = models.CharField(max_length=96)

    site = models.ForeignKey(Site, editable=False, default=settings.SITE_ID)
    on_site = CollectionManager('site')

    def __unicode__(self):
        return u"%r (user: %r, site: %r)" % (self.name, self.user.username, self.site)

    class Meta:
        ordering = ("-modified",)


class WboManager(models.Manager):
    def create_or_update(self, payload_dict, collection, user, timestamp):
        """
        create or update a wbo
        TODO:
            - Check parentid, but how?
            - must wboid + parentid be unique?
        """
        wbo, created = Wbo.objects.get_or_create(
            collection=collection,
            user=user,
            wboid=payload_dict['id'],
            defaults={
                'parentid': payload_dict.get('parentid', None),
                'predecessorid': payload_dict.get('predecessorid', None),
                'sortindex': payload_dict.get('sortindex', None),
                'modified': timestamp,
                'payload': payload_dict['payload'],
            }
        )
        if not created:
            wbo.parentid = payload_dict.get("parentid", None)
            wbo.predecessorid = payload_dict.get("predecessorid", None)
            wbo.sortindex = payload_dict.get("sortindex", None)
            wbo.modified = timestamp,
            wbo.payload = payload_dict["payload"]
            wbo.save()

        return wbo, created


class Wbo(BaseModel):
    """
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API   
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/Setup
    
    inherited from BaseModel:
        modified -> datetime of the last change
    """
    objects = WboManager()
    collection = models.ForeignKey(Collection, blank=True, null=True)
    user = models.ForeignKey(User)
    wboid = models.CharField(max_length=64,
        help_text="wbo identifying string"
    )
    parentid = models.CharField(max_length=64, blank=True, null=True,
        help_text="wbo parent identifying string"
    )
    predecessorid = models.CharField(max_length=64, blank=True, null=True,
        help_text="wbo predecessorid"
    )
    sortindex = models.IntegerField(blank=True, null=True,
        help_text="An integer indicting the relative importance of this item in the collection."
    )
    payload = models.TextField(blank=True,
        help_text=(
            "A string containing a JSON structure encapsulating the data of the record."
            " This structure is defined separately for each WBO type. Parts of the"
            " structure may be encrypted, in which case the structure should also"
            " specify a record for decryption."
        )
    )

    def get_response_dict(self):
        response_dict = {
            "id": self.wboid,
            "modified": weave_timestamp(self.modified),
            "payload": self.payload,
        }
        # Don't send additional properties if payload is emtpy -> deleted Wbo
        if self.payload != '':
            for key in ("parentid", "predecessorid", "sortindex"):
                value = getattr(self, key)
                if value:
                    response_dict[key] = value
        return response_dict

    def __unicode__(self):
        return u"%r (%r)" % (self.wboid, self.collection)

    class Meta:
        ordering = ("-modified",)



#coding:utf-8

from django.shortcuts import render_to_response
from django.core.urlresolvers import reverse
from django.contrib.auth.decorators import login_required

def absolute_uri(request, view_name, **kwargs):
    url = reverse(view_name, kwargs=kwargs)
    absolute_uri = request.build_absolute_uri(url)
    return absolute_uri


@login_required
def url_info(request):
    context = {
        "title": "weave testproject url info",
        "server_url": request.build_absolute_uri("weave"),
        "register_check_url": absolute_uri(request, "weave-register_check", username=request.user.username),
        "info_url": absolute_uri(request, "weave-info", version="1.0", username=request.user.username),
    }

    return render_to_response("testproject/url_info.html", context)



'''
Register all weave models in Django admin interface.

Created on 15.03.2010

@license: GNU GPL v3 or above, see LICENSE for more details.
@copyright: 2010 see AUTHORS for more details.
@author: Jens Diemer
@author: Michael Fladischer <michael@fladi.at>
'''

from django.contrib import admin
# TODO: Find a way to incorporate django-reversion
#from reversion.admin import VersionAdmin

from weave.models import Wbo, Collection

class WboAdminInline(admin.TabularInline):
    model = Wbo

class WboAdmin(admin.ModelAdmin):
    def payload_cutout(self, obj):
        MAX = 100
        payload = obj.payload
        if len(payload) > MAX:
            payload = payload[:MAX] + "..."
        return payload
    payload_cutout.short_description = "Payload cutout"
    list_display = ['id', "user", "wboid", 'collection', "parentid", "modified", "sortindex", "payload_cutout"]
    list_filter = ['user', 'collection']
    date_hierarchy = 'modified'
    search_fields = ("wboid", "parentid", "sortindex", "payload")

admin.site.register(Wbo, WboAdmin)

class CollectionAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'user', 'modified', 'site']
    list_filter = ['user']
    date_hierarchy = 'modified'
    inlines = [WboAdminInline]

admin.site.register(Collection, CollectionAdmin)



# coding: utf-8

'''
    Storage for Weave.
    
    Created on 15.03.2010
    
    @license: GNU GPL v3 or above, see LICENSE for more details.
    @copyleft: 2010 by the django-weave team, see AUTHORS for more details.
'''

from datetime import datetime

try:
    import json # New in Python v2.6
except ImportError:
    from django.utils import simplejson as json

from django.contrib.csrf.middleware import csrf_exempt
from django.shortcuts import get_object_or_404
from django.core.exceptions import ValidationError

# django-weave own stuff
from weave.models import Collection, Wbo
from weave.utils import limit_wbo_queryset, weave_timestamp, assert_weave_version
from weave.decorators import weave_assert_username, logged_in_or_basicauth, \
  weave_render_response
from weave import Logging

logger = Logging.get_logger()

@logged_in_or_basicauth
@weave_assert_username
@csrf_exempt
@weave_render_response
def info(request, version, username, timestamp):
    """
    return all collection keys with the timestamp
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#GET
    """
    assert_weave_version(version)
    collections = dict()
    for collection in Collection.on_site.filter(user=request.user):
        collections[collection.name] = weave_timestamp(collection.modified)
    return collections


@logged_in_or_basicauth
@weave_assert_username
@csrf_exempt
@weave_render_response
def storage(request, version, username, timestamp, col_name=None, wboid=None):
    """
    Handle storing Collections and WBOs. 
    """
    assert_weave_version(version)
    if 'X-If-Unmodified-Since' in request.META:
        since = datetime.fromtimestamp(float(request.META['X-If-Unmodified-Since']))
    else:
        since = None

    if request.method == 'GET':
        # Returns a list of the WBO contained in a collection.
        collection = get_object_or_404(Collection.on_site, user=request.user, name=col_name)

        if wboid is not None: # return one WBO
            wbo = get_object_or_404(Wbo, user=request.user, collection=collection, wboid=wboid)
            return wbo.get_response_dict()

        wbo_queryset = Wbo.objects.filter(user=request.user, collection=collection)
        wbo_queryset = limit_wbo_queryset(request, wbo_queryset)

        # If defined, returns the full WBO, rather than just the id. 
        if not 'full' in request.GET: # return only the WBO ids
            return [wbo.wboid for wbo in wbo_queryset]

        return [wbo.get_response_dict() for wbo in wbo_queryset]

    elif request.method == 'PUT':
        # https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#PUT
        # Adds the WBO defined in the request body to the collection.

        payload = request.raw_post_data
        logger.debug("Payload for PUT: %s" % payload)
        if not payload:
            raise NotImplementedError

        val = json.loads(payload)

        if val.get('id', None) != wboid:
            raise ValidationError

        # TODO: I don't think that it's good to just pass 0 in case the header is not defined
        collection, created = Collection.on_site.create_or_update(
                                                                  request.user,
                                                                  col_name,
                                                                  timestamp,
                                                                  since,
                                                                  )

        if created:
            logger.debug("Created new collection %s" % collection)
        else:
            logger.debug("Found existing collection %s" % collection)

        wbo, created = Wbo.objects.create_or_update(val, collection, request.user, timestamp)

        if created:
            logger.debug("New wbo created: %r" % wbo)
        else:
            logger.debug("Existing wbo updated: %r" % wbo)

        return weave_timestamp(timestamp)

    elif request.method == 'POST':
        status = {'success': [], 'failed': []}
        payload = request.raw_post_data
        logger.debug("Payload in POST: %s" % payload)
        if not payload:
            raise NotImplementedError

        # TODO: I don't think that it's good to just pass 0 in case the header is not defined
        collection, created = Collection.on_site.create_or_update(
                                                                  request.user,
                                                                  col_name,
                                                                  timestamp,
                                                                  since,
                                                                  )

        if created:
            logger.debug("Create new collection %s" % collection)
        else:
            logger.debug("Found existing collection %s" % collection)

        data = json.loads(payload)

        if not isinstance(data, list):
            raise NotImplementedError

        for item in data:
            wbo, created = Wbo.objects.create_or_update(item, collection, request.user, timestamp)
            status['success'].append(wbo.wboid)

            if created:
                logger.debug("New wbo created: %s" % wbo)
            else:
                logger.debug("Existing wbo updated: %s" % wbo)

        return status
    elif request.method == 'DELETE':
        # FIXME: This is am mess, it needs better structure
        if col_name is None and wboid is None:
            assert request.META.has_key('X-Confirm-Delete'), "To remove all records for your user, please make sure to include a X-Confirm-Delete HTTP header in your DELTE request!"
            Collection.on_site.filter(user=request.user).delete()
            return weave_timestamp(timestamp)

        if col_name is not None and wboid is not None:
            wbo = Wbo.objects.get(user=request.user, collection__name=col_name, wboid=wboid)
            if wbo is not None:
                logger.info("Delete Wbo %s in collection %s for user %s" % (wbo.wboid, col_name, request.user))
                wbo.delete()
            else:
                logger.info("Deletion of wboid %s requested but there is no such wbo!" % (wboid))
        else:
            ids = request.GET.get('ids', None)
            if ids is not None:
                for wbo in Wbo.objects.filter(user=request.user, wboid__in=ids.split(',')):
                    logger.info("Delete Wbo %s in collection %s for user %s" % (wbo.wboid, col_name, request.user))
                    wbo.delete()
            else:
                collection = Collection.on_site.filter(user=request.user, name=col_name).delete()
                if collection is not None:
                    logger.info("Delete collection %s for user %s" % (collection.name, request.user))
                    Wbo.objects.filter(user=request.user, collection=collection).delete()
                else:
                    logger.info("Deletion of collection %s requested but there is no such collection!" % (col_name))
        return weave_timestamp(timestamp)

    else:
        raise NotImplementedError("request.method %r not implemented!" % request.method)



# coding: utf-8

from django.conf.urls.defaults import patterns, include, url

from django.contrib import admin
admin.autodiscover()

import weave.urls
from testproject.views import url_info


handler404 = 'django.views.defaults.page_not_found'
handler500 = 'django.views.defaults.server_error'


urlpatterns = patterns('',
    url(r'^$', url_info),
    url(r'^weave/', include(weave.urls), name="weave-root"),

    url(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    url(r'^admin/', include(admin.site.urls)),
)



# coding:utf-8

'''
    Decorators for view functions inside weave.
    HTTP basic auth decorators taken from:
    http://www.djangosnippets.org/snippets/243/

    Info:
    ~~~~~
    to debug the output in a browser: add "?debug=1" to a url.
    To reformat the payload: add "?debug=2" to a url.
    (This works only if settings.DEBUG==True)

    Created on 15.03.2010
    
    @license: GNU GPL v3 or above, see LICENSE for more details.
    @copyright: 2010 see AUTHORS for more details.
    @author: Jens Diemer
    @author: Scanner (http://www.djangosnippets.org/users/Scanner/)
    @author: Michael Fladischer <michael@fladi.at>
'''

import base64
from datetime import datetime

try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps # Python 2.3, 2.4 fallback.

try:
    import json # New in Python v2.6
except ImportError:
    from django.utils import simplejson as json

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.contrib.auth import authenticate, login

from weave.utils import weave_timestamp
from weave import Logging

logger = Logging.get_logger()

def view_or_basicauth(view, request, test_func, realm="", *args, **kwargs):
    """
    This is a helper function used by both 'logged_in_or_basicauth' and
    'has_perm_or_basicauth' that does the nitty of determining if they
    are already logged in or if they have provided proper http-authorization
    and returning the view if all goes well, otherwise responding with a 401.
    """
    if test_func(request.user):
        # Already logged in, just return the view.
        return view(request, *args, **kwargs)

    # They are not logged in. See if they provided login credentials
    if 'HTTP_AUTHORIZATION' in request.META:
        auth = request.META['HTTP_AUTHORIZATION'].split()
        if len(auth) == 2:
            # NOTE: We are only support basic authentication for now.
            #
            if auth[0].lower() == "basic":
                uname, passwd = base64.b64decode(auth[1]).split(':')
                user = authenticate(username=uname, password=passwd)
                if user is not None:
                    if user.is_active:
                        login(request, user)
                        request.user = user
                        return view(request, *args, **kwargs)

    # Either they did not provide an authorization header or
    # something in the authorization attempt failed. Send a 401
    # back to them to ask them to authenticate.
    response = HttpResponse()
    response.status_code = 401
    response['WWW-Authenticate'] = 'Basic realm="%s"' % realm
    return response

def logged_in_or_basicauth(func, realm=None):
    """
    A simple decorator that requires a user to be logged in. If they are not
    logged in the request is examined for a 'authorization' header.
    
    If the header is present it is tested for basic authentication and
    the user is logged in with the provided credentials.
    
    If the header is not present a http 401 is sent back to the
    requestor to provide credentials.
    
    The purpose of this is that in several django projects I have needed
    several specific views that need to support basic authentication, yet the
    web site as a whole used django's provided authentication.
    
    The uses for this are for urls that are access programmatically such as
    by rss feed readers, yet the view requires a user to be logged in. Many rss
    readers support supplying the authentication credentials via http basic
    auth (and they do NOT support a redirect to a form where they post a
    username/password.)
    
    Use is simple:
    
    @logged_in_or_basicauth
    def your_view:
        ...
    
    You can provide the name of the realm to ask for authentication within.
    """
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        return view_or_basicauth(func, request,
                                 lambda u: u.is_authenticated(),
                                 realm, *args, **kwargs)
    return wrapper

def has_perm_or_basicauth(func, perm, realm=""):
    """
    This is similar to the above decorator 'logged_in_or_basicauth'
    except that it requires the logged in user to have a specific
    permission.
    
    Use:
    
    @logged_in_or_basicauth('asforums.view_forumcollection')
    def your_view:
        ...
    
    """
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        return view_or_basicauth(func, request,
                                 lambda u: u.has_perm(perm),
                                 realm, *args, **kwargs)
    return wrapper

def weave_assert_username(func, key='username'):
    """
    Decorator to check if the username from the URL is the one logged in.
    It uses the kwargs "username" as the default key but it can be changed by passing
    key='field' to the decorator. 
    
    Use:
    
    @assert_username
    def your_view(request, username):
    
    You can provide the key of the username field which is 'username' by default.
    """
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        # Test if username argument matches logged in user.
        # Weave uses lowercase usernames inside the URL!!!
        if request.user.username != kwargs[key]:
            logger.debug("Logged in user %s does not match %s from URL." % (request.user.username, kwargs[key]))
            raise PermissionDenied("Username from HTTP authentication does not match URL!")
        return func(request, *args, **kwargs)
    return wrapper

def weave_render_response(func):
    """
    Decorator that checks for the presence of weave specific HTTP headers 
    and formats output accordingly.
    
    Use:
    
    @weave_render_response
    def your_view(request):
    """
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        timedata = datetime.now()
        data = func(request, timestamp=timedata, *args, **kwargs)
        response = HttpResponse()
        response["X-Weave-Timestamp"] = weave_timestamp(timedata)

        if settings.DEBUG and "debug" in request.GET:
            logger.debug("debug output for %r:" % func.__name__)

            if int(request.GET["debug"]) > 1:
                def load_payload(item):
                    if "payload" in item:
                        raw_payload = item["payload"]
                        payload_dict = json.loads(raw_payload)
                        item["payload"] = payload_dict
                    return item

                if isinstance(data, list):
                    data = [load_payload(item) for item in data]
                else:
                    data = load_payload(data)

            response["content-type"] = "text/plain"
            response.content = json.dumps(data, indent=4)
        else:
            if request.META['HTTP_ACCEPT'] == 'application/newlines' and isinstance(data, list):
                response.content = '\n'.join([json.dumps(element) for element in data]) + '\n'
                response["content-type"] = 'application/newlines'
                response['X-Weave-Records'] = len(data)
            else:
                response["content-type"] = "application/json"
                response.content = json.dumps(data)

        return response
    return wrapper



# coding:utf-8

'''
    User handling for Weave.
    FIXME: Not complete yet.
    
    Created on 15.03.2010
    
    @license: GNU GPL v3 or above, see LICENSE for more details.
    @copyleft: 2010 by the django-weave team, see AUTHORS for more details.
'''
try:
    import json # New in Python v2.6
except ImportError:
    from django.utils import simplejson as json

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.contrib.auth.models import User
from django.contrib.csrf.middleware import csrf_exempt
from django.http import HttpResponseBadRequest, HttpResponse, \
    HttpResponseNotFound

# django-weave own stuff
from weave import constants
from weave.decorators import logged_in_or_basicauth
from weave import Logging
from weave.utils import assert_weave_version

logger = Logging.get_logger()

@logged_in_or_basicauth
@csrf_exempt
def password(request):
    """
    Change the user password.
    """
    if request.method != 'POST':
        logger.error("wrong request method %r" % request.method)
        return HttpResponseBadRequest()

    # Make sure that we are able to change the password. 
    # If for example django-auth-ldap is used for authentication it will set the password for 
    # User objects to a unusable one in the database. Therefore we cannot change it, it has to 
    # happen inside LDAP.
    if not request.user.has_usable_password():
        logger.debug("Can't change password. User %s has a unusable password." % request.user.username)
        return HttpResponseBadRequest()

    # The PHP server for Weave uses the first 2048 (if there is enough data) characters
    # from POST data as the new password. We decided to throw an error if the password 
    # data is longer than 256 characters.
    if len(request.raw_post_data) > 256:
        msg = (
            "Don't change password for user %s."
            " POST data has more than 256 characters! (len=%i)"
        ) % (request.user.username, len(request.raw_post_data))
        logger.debug(msg)
        return HttpResponseBadRequest()

    request.user.set_password(request.raw_post_data)
    request.user.save()
    logger.debug("Password for User %r changed to %r" % (request.user.username, request.raw_post_data))
    return HttpResponse()


@csrf_exempt
def node(request, version, username):
    """
    finding cluster for user -> return 404 -> Using serverURL as data cluster (multi-cluster support disabled)
    """
    assert_weave_version(version)

    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        logger.debug("User %r doesn't exist!" % username)
        return HttpResponseNotFound(constants.ERR_UID_OR_EMAIL_AVAILABLE)
    else:
        logger.debug("User %r exist." % username)
        #FIXME: Send the actual cluster URL instead of 404
        return HttpResponseNotFound(constants.ERR_UID_OR_EMAIL_IN_USE)


@csrf_exempt
def register_check(request, username):
    """
    Returns 1 if the username exist, 0 if not exist.
    """
    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        logger.debug("User %r doesn't exist!" % username)
        return HttpResponse(constants.ERR_UID_OR_EMAIL_AVAILABLE)
    else:
        logger.debug("User %r exist." % username)
        return HttpResponse(constants.ERR_UID_OR_EMAIL_IN_USE)


@csrf_exempt
def exists(request, version, username):
    """
    https://wiki.mozilla.org/Labs/Weave/User/1.0/API#GET
    Returns 1 if the username is in use, 0 if it is available.
    
    e.g.: https://auth.services.mozilla.com/user/1/UserName
    """
    if request.method == 'GET':
        try:
            User.objects.get(username=username)
        except User.DoesNotExist:
            logger.debug("User %r doesn't exist!" % username)
            return HttpResponse("0")
        else:
            logger.debug("User %r exist." % username)
            return HttpResponse("1")
    elif request.method == 'PUT':
        # Check for aviability of recaptcha 
        # (can be found at: http://pypi.python.org/pypi/recaptcha-client)
        try:
            from recaptcha.client.captcha import submit
        except ImportError:
            logger.error("Captcha requested but unable to import the 'recaptcha' package!")
            return HttpResponse("Captcha support disabled due to missing Python package 'recaptcha'.")
        if not getattr(settings, "RECAPTCHA_PRIVATE_KEY"):
            logger.error("Trying to create user but settings.RECAPTCHA_PRIVATE_KEY not set")
            raise ImproperlyConfigured
        # Handle user creation.
        data = json.loads(request.raw_post_data)
        # Usernames are limited to a length of max. 30 chars.
        # http://docs.djangoproject.com/en/dev/topics/auth/#django.contrib.auth.models.User.username
        if len(username) > 30 or len(data['password']) > 256:
            return HttpResponseBadRequest()
        result = submit(
                        data['captcha-challenge'],
                        data['captcha-response'],
                        settings.RECAPTCHA_PRIVATE_KEY,
                        request.META['REMOTE_ADDR']
                        )
        if not result.is_valid:
            # Captcha failed.
            return HttpResponseBadRequest()
        User.objects.create_user(username, data['email'], data['password'])
        return HttpResponse()
    else:
        raise NotImplemented()
